/* xpiLinks.js
 * Queries remote update.rdf to find the links to the XPIs installed on this system
 * Requires: common.js, addonsProvider.js, collect.js
 *
 * Part of InfoLister by Nickolay Ponomarev <asqueella@gmail.com>
 * Copyright (c) 2004-2005, Nickolay Ponomarev.
 * See the license.txt included in this package for licensing information.
 *
 * Initial version of this code has been contributed by "zeniko".
 */

var appID, appVersion;

var itemsToProcess;
var xpiLinksCacheOld = {};

// Collect links to XPI files for each extension and theme in addons.
// addons is an array or NodeList of DOM nodes for items.
function collectXPILinks(addons, setProgressCallback, doneCallback) {
  LOG_xpiLinks("== collectXPILinks start ==");

  xpiLinksCache = {};
  xpiLinksCacheOld = {};
  try {
    xpiLinksCacheOld = eval(ILPrefs.getUnicharPrefDef("xpiLinksCache", "({})"));
  } catch(e) {
    Components.reportError(e);
    ILPrefs.clearUserPref("xpiLinksCache");
  }

  xpiLinksProgress.init(addons.length, setProgressCallback, doneCallback);
  if(addons.length == 0) {
    xpiLinksProgress.done();
    return;
  }
  itemsToProcess = [];
  for(var i = 0; i < addons.length; i++)
    itemsToProcess.push(addons.item(i)); // xxx for some reason firefox doesn't like addons[i] here (issues a strict warning) :-/

// process all items simultaneously: i = itemsToProcess.length-1; i >= 0; i--) {
  for(var i = 0; i < 2; i++) // process 2 items at a time
    processNextItem();        // - this will call processNextItem again when done
}


function processNextItem() {
  if(itemsToProcess.length == 0)
    return false;
  var addon = itemsToProcess.pop();
  var itemID = addon.getAttribute("id");
  xpiLinksProgress.startProcessing(itemID);

  if(itemID in xpiLinksCacheOld) { // cached?
    xpiLinksCache[itemID] = xpiLinksCacheOld[itemID];
    xpiLinksProgress.doneProcessing(itemID);
    return true;
  }
  if(addon.getAttribute("appManaged") == "true") {
    // XXX  the XPIs for app-managed items cannot be found via update mechanism
    xpiLinksProgress.doneProcessing(itemID);
    return true;
  }

  try {
    // ---- cf. RDFItemUpdater.checkForUpdates() in nsExtensionManager.js ----

    // Look for a custom update URI: 1) supplied by a pref, 2) supplied by the
    // install manifest, 3) the default configuration
    dsURI = null;
    try {
      var dsURI = ILPrefs.prefSvc.getComplexValue(
        "extensions.%UUID%.update.url".replace(/%UUID%/, itemID), 
        CI.nsIPrefLocalizedString).data;
    } catch (e) {}

    if (!dsURI) {
      dsURI = addon.getAttribute("updateURL");
      LOG_xpiLinks("getting updateURL for " + itemID + ": " + dsURI)
    } if (!dsURI) {
      dsURI = ILPrefs.prefSvc.getComplexValue("extensions.update.url",
        CI.nsIPrefLocalizedString).data;
    }

    var xulappInfo = ILHelpers.getService("xre/app-info;1", "nsIXULAppInfo");
    appID = xulappInfo.ID.toString();
    appVersion = xulappInfo.version;

    // we cheat on %ITEM_MAXAPPVERSION%, and %ITEM_VERSION%
    // the rest is as in nsRdfItemUpdater.checkForUpdates
    dsURI = dsURI.replace(/%ITEM_ID%/g, itemID);
    dsURI = dsURI.replace(/%ITEM_VERSION%/g, "0.0.0");
    dsURI = dsURI.replace(/%ITEM_MAXAPPVERSION%/g, appVersion);
    dsURI = dsURI.replace(/%APP_ID%/g, appID);
    dsURI = dsURI.replace(/%APP_VERSION%/g, appVersion);
    dsURI = dsURI.replace(/%REQ_VERSION%/g, 1);
    dsURI = dsURI.replace(/%APP_OS%/g, ILHelpers.appInfo.OS);
    try {
      dsURI = dsURI.replace(/%APP_ABI%/g, ILHelpers.appInfo.XPCOMABI);
    } catch(e) {
      // XPCOM ABI information might not be available
    }

    LOG_xpiLinks("reading update.rdf at " + dsURI);
    var ds = gRDF.GetDataSource(dsURI);
    ds.QueryInterface(CI.nsIRDFRemoteDataSource)
    if(ds.loaded)
      onDatasourceLoaded(ds, itemID);
    else {
      var sink = ds.QueryInterface(CI.nsIRDFXMLSink);
      sink.addXMLSinkObserver(new UpdateRDFSinkObserver(itemID));
    }
  } catch(e) {
    LOG_xpiLinks("There was an error loading the update datasource at URI=" + 
      dsURI + ", \r\nitemID=" + itemID + "\r\nerror: " +
      ("message" in e ? e.message : e));
    xpiLinksProgress.doneProcessing(itemID);
  }
  return true;
}

/**
 * This function is called when an update manifest is finished loading, whether
 * the loading was synchronous or asynchronous.
 * @param {nsIRDFDataSource} datasource The datasource object for the manifest.
 * @param {string} itemID
 */
function onDatasourceLoaded(datasource, itemID) {
  var xpiLink = getXPILink(datasource, itemID);
  LOG_xpiLinks(itemID + ": " + xpiLink);
  if(xpiLink)
    xpiLinksCache[itemID] = xpiLink;
  xpiLinksProgress.doneProcessing(itemID);
}

/**
 * An instance of this class is created for each item, whose update manifest is 
 * loaded asynchronously.
 */
function UpdateRDFSinkObserver(itemID) {
  this.itemID = itemID;
}

UpdateRDFSinkObserver.prototype = {
  _removeXMLSinkObserver: function(aSink) {
    try {
      aSink.removeXMLSinkObserver(this);
    }
    catch (e) {
      LOG_xpiLinks("updateRDFSinkObserver:onError: Failure during removeXMLSinkObserver call");
    }
  },

  // nsIRDFXMLSinkObserver
  onBeginLoad: function(aSink) {
  },
  onInterrupt: function(aSink) {
  },
  onResume: function(aSink) {
  },
  onEndLoad: function(aSink) {
    this._removeXMLSinkObserver(aSink);
    try {
      onDatasourceLoaded(aSink.QueryInterface(CI.nsIRDFDataSource), this.itemID);
    } catch (e) { 
      LOG_xpiLinks("updateRDFSinkObserver:onEndLoad " + e);
      xpiLinksProgress.doneProcessing(this.itemID);
    }
  },
  onError: function(aSink, aStatus, aErrorMsg) {
    LOG_xpiLinks("updateRDFSinkObserver:onError: There was an error loading the \r\n" + 
      "the update datasource for item " + this.itemID + ", status=" + aStatus + 
      "\r\nerror=" + aErrorMsg);
    xpiLinksProgress.doneProcessing(this.itemID);
    this._removeXMLSinkObserver(aSink);
  }
};

// gets a link to the XPI given an RDF datasource object for the update manifest
// and the item's ID.
function getXPILink(datasource, itemID) {
  var XMLNS_EM = "http://www.mozilla.org/2004/em-rdf#";
  var extensionRes = gRDF.GetResource("urn:mozilla:extension:" + itemID); // XXX
  var updates = datasource.GetTarget(extensionRes, gRDF.GetResource(XMLNS_EM + "updates"), true);

  if (!updates)
  { // maybe we found an old update.rdf (for Firefox v0.9)
    var updateLinkResource = datasource.GetTarget(extensionRes, gRDF.GetResource(XMLNS_EM + "updateLink"), true);
    return updateLinkResource ? updateLinkResource.QueryInterface(CI.nsIRDFLiteral).Value : "";
  }

  var updatesCont = CC["@mozilla.org/rdf/container;1"].createInstance(CI.nsIRDFContainer);
  updatesCont.Init(datasource, updates.QueryInterface(CI.nsIRDFResource));

  // for simplicity, we take the first version we find (it's usually the one we want)
  var versionInfo = updatesCont.GetElements().getNext().QueryInterface(CI.nsIRDFResource);
  var targetApps = datasource.GetTargets(versionInfo, gRDF.GetResource(XMLNS_EM + "targetApplication"), true);
  while (targetApps.hasMoreElements())
  {
    var targetAppInfo = targetApps.getNext().QueryInterface(CI.nsIRDFResource);
    var targetAppID = datasource.GetTarget(targetAppInfo, gRDF.GetResource(XMLNS_EM + "id"), true).QueryInterface(CI.nsIRDFLiteral).Value;
    if (targetAppID != appID)
      continue;

    var updateLink = datasource.GetTarget(targetAppInfo, gRDF.GetResource(XMLNS_EM + "updateLink"), true);
    if (updateLink)
      return updateLink.QueryInterface(CI.nsIRDFLiteral).Value;
  }
  return "";
}

/**
 * This object is responsible for managing progress info.
 */
var xpiLinksProgress = {
  // An array of IDs of items being processed
  processing: [],
  processedAddons: 0,
  totalAddons: 0,
  cancelled: false,

  init: function(totalAddons, setProgressCallback, doneCallback) {
    this.totalAddons = totalAddons;
    this.setProgressCallback = setProgressCallback;
    this.doneCallback = doneCallback;
    this.setProgressCallback(0);
  },

  cancel: function() {
    this.cancelled = true;
    this.done();
  },

  startProcessing: function(itemID) {
    LOG_xpiLinks("started processing " + itemID);
    this.processing.push(itemID);
  },

  doneProcessing: function(itemID) {
    LOG_xpiLinks("done with " + itemID);
    var idx = this.processing.indexOf(itemID);
    if(idx == -1) {
      LOG_xpiLinks("oops! " + itemID + " is not in processing="+this.processing.toSource());
      return;
    }

    this.processing[idx] = this.processing[this.processing.length-1];
    -- this.processing.length;

    if(!this.cancelled)
      this.setProgressCallback(100 * (++this.processedAddons) / this.totalAddons);
    if(this.processedAddons == this.totalAddons) {
      if(this.processing.length != 0)
        LOG_xpiLinks("oops! processing=" + this.processsing.toSource() + ", but processedAddons=totalAddons");
      this.done();
    } else if(!this.cancelled)
      processNextItem();
  },

  done: function() {
    LOG_xpiLinks("collectXPILinks end: " + xpiLinksCache.toSource());
    ILPrefs.setUnicharPref("xpiLinksCache", xpiLinksCache.toSource());
    if(this.doneCallback)
      this.doneCallback(this.cancelled);
  }
};


function LOG_xpiLinks(aMsg) {
  //LOG(aMsg);
}
